// Copyright (c) 2010 Sundog Software, LLC. All rights reserved worldwide.

#include "stdafx.h"
#include "nvsg/GeoNode.h"
#include "nvsg/Camera.h"
#include "nvsg/DirectedLight.h"
#include "nvmath/Spherent.h"
#include "SilverLiningManager.h"

using namespace nvsg;
using namespace nvutil;
using namespace SilverLining;

typedef struct SilverLiningData_S
{
    SilverLining::Atmosphere *atm;
    CameraHandle *camera;
    DirectedLightHandle *light;
} SilverLiningData;

Atmosphere *SilverLiningManager::AttachToRoot(GroupHandle *rootNode, CameraHandle *camera,
                                              DirectedLightHandle *light)
{
    // Create an atmosphere object:
    Atmosphere *atm = new Atmosphere("Your user name", "Your license key");

    const char *slPath = getenv("SILVERLINING_PATH");
    std::string resourcePath(slPath);
    resourcePath += "\\resources\\";

    int result = atm->Initialize(Atmosphere::OPENGL, resourcePath.c_str(), true, NULL);
    if (result != Atmosphere::E_NOERROR)
    {
        return NULL;
    }

    // Populate data we want to have access to
    SilverLiningData *slData = new SilverLiningData;
    slData->atm = atm;
    slData->camera = camera;
    slData->light = light;

    // Attach the sky at the beginning:
    SkyHandle *hdlSky = CreateHandle(SkyHandle);
    WritableObject<SilverLiningSky> theSky(hdlSky);
    theSky->setUserData(slData);
    theSky->setDrawCallback(new SilverLiningSkyDraw(atm));

    GeoNodeHandle *hdlSkyNode = CreateHandle(GeoNodeHandle);
    WritableObject<GeoNode> theSkyNode(hdlSkyNode);
    theSkyNode->addGeometry(hdlSky);
    theSkyNode->addHints(Object::NVSG_HINT_ALWAYS_VISIBLE);

    WritableObject<Group> theGroup(rootNode);
    theGroup->addChild(hdlSkyNode, 0);

    // Attach the clouds at the end:
    CloudsHandle *hdlClouds = CreateHandle(CloudsHandle);
    WritableObject<SilverLiningClouds> theClouds(hdlClouds);
    theClouds->setUserData(slData);
    theClouds->setDrawCallback(new SilverLiningCloudsDraw(atm));

    GeoNodeHandle *hdlCloudNode = CreateHandle(GeoNodeHandle);
    WritableObject<GeoNode> theCloudNode(hdlCloudNode);
    theCloudNode->addGeometry(hdlClouds);

    theGroup->addChild(hdlCloudNode);

    return atm;
}

void SilverLiningSkyDraw::updateLight(DirectedLightHandle *lightHdl)
{
    if (!lightHdl || !atmosphere) return;

    float rd, gd, bd; // directional light
    float ra, ga, ba; // ambient light
    float x, y, z;

    atmosphere->GetSunOrMoonColor(&rd, &gd, &bd);
    atmosphere->GetAmbientColor(&ra, &ga, &ba);
    atmosphere->GetSunOrMoonPosition(&x, &y, &z);

    WritableObject<DirectedLight> light(lightHdl);
    light->setEnabled(true);
    light->setAmbientColor(nvmath::Vec3f(ra, ga, ba));
    light->setDiffuseColor(nvmath::Vec3f(rd, gd, bd));
    light->setSpecularColor(nvmath::Vec3f(rd, gd, bd));
    light->setDirection(nvmath::Vec3f(-x, -y, -z));
}

void SilverLiningSkyDraw::extractMatrices(CameraHandle *camHdl)
{
    if (!camHdl || !atmosphere) return;

    const Camera *cam = ReadableObject<Camera>(camHdl);
    nvmath::Mat44f viewMat = cam->getWorldToViewMatrix();
    const float *fViewMat = viewMat.getPtr();
    double dCamMat[16];
    for (int i = 0; i < 16 ;i++)
    {
        dCamMat[i] = fViewMat[i];
    }
    atmosphere->SetCameraMatrix(dCamMat);   

    nvmath::Mat44f projMat = cam->getProjection();
    const float *fProjMat = projMat.getPtr();
    double dProjMat[16];
    for (int j = 0; j < 16; j++)
    {
        dProjMat[j] = fProjMat[j];
    }
    atmosphere->SetProjectionMatrix(dProjMat);
}

bool SilverLiningSkyDraw::draw(const Drawable *drawable)
{
    SilverLiningData *slData = (SilverLiningData *)drawable->getUserData();
    if (slData)
    {
        // Extract the view and projection matrices and pass them to SilverLining:
        extractMatrices(slData->camera);

        // Draw the sky box:
        atmosphere->BeginFrame(true);

        // Sync up the directional light:
        updateLight(slData->light);
    }

    // Tell SceniX we handled the drawing ourselves:
    return true;
}

bool SilverLiningCloudsDraw::draw(const Drawable *drawable)
{
    // Draw the clouds:
    atmosphere->EndFrame();
    
    // Tell SceniX we handled the drawing ourselves:
    return true;
}

SilverLiningSky::~SilverLiningSky()
{
    SilverLiningData *slData = (SilverLiningData *)getUserData();
    if (slData)
    {
        delete slData->atm;
        delete slData;
    }
}

nvmath::Sphere3f  SilverLiningSky::calculateBoundingSphere () const
{
    // Compute the size and position of the sky box.

    SilverLiningData *slData = (SilverLiningData *)getUserData();
    float skyBoxRadius = 2000.0f;
    nvmath::Vec3f camPos(0, 0, 0);

    if (slData)
    {
        double boxSize = slData->atm->GetConfigOptionDouble("sky-box-size");
        skyBoxRadius = (float)boxSize * 1.42f;

        CameraHandle *camHandle = slData->camera;
        camPos = ReadableObject<Camera>(camHandle)->getPosition();
    }

    nvmath::Sphere3f s;
    s.setCenter(camPos);
    s.setRadius(skyBoxRadius);

    return s;
}

nvmath::Sphere3f  SilverLiningClouds::calculateBoundingSphere () const
{
    // Compute the size and position of all cloud layers.
    // Since we're only using camera-centered "infinite" cloud layers, this is 
    // fairly simple. If you had non-infinite cloud layers, you'd need to 
    // intersect the bounds of each layer here taking their positions into account.

    SilverLiningData *slData = (SilverLiningData *)getUserData();

    nvmath::Vec3f cloudsCenter(0, 0, 0);
    float cloudsRadius = 100000;

    if (slData)
    {
        nvmath::Vec3f camPos = ReadableObject<Camera>(slData->camera)->getPosition();
        cloudsCenter = camPos;
        cloudsCenter[1] = 0; // Note, we're assuming the default y-is-up convention here! Change this otherwise.

        Atmosphere *atm = slData->atm;
        std::map<int, CloudLayer*>& cloudLayers = atm->GetConditions()->GetCloudLayers();
        std::map<int, CloudLayer*>::iterator it;

        float maxRadius = 0;

        for (it = cloudLayers.begin(); it != cloudLayers.end(); it++)
        {
            int cloudHandle = it->first;
            CloudLayer *layer = it->second;

            double length = layer->GetBaseLength();
            double width = layer->GetBaseWidth();
            double thickness = layer->GetThickness();
            double altitude = layer->GetBaseAltitude();
        
            double yr = thickness + altitude;
            double xr = width * 0.5;
            double zr = length * 0.5;
            double radius = sqrt(xr*xr + yr*yr + zr*zr);
            
            if (radius > maxRadius)
            {
                maxRadius = (float)radius;
            }
        }

        cloudsRadius = maxRadius;
    }

    nvmath::Sphere3f s;
    s.setCenter(cloudsCenter);
    s.setRadius(cloudsRadius);

    return s;
}
